#include <iostream>
#include <unistd.h>
#include <cstring>
#include <string>
#include <sstream>
#include <iomanip>
#include <vector>
#include <sys/types.h>
#include <sys/wait.h>

//########################################
//##  Gestion des erreurs
//########################################
/**
   @brief teste errnum et envoie une exception si errnum est un code d'erreur différent de 0
*/
#define exit_error(msg, cond, errnum) __error_(__FILE__, __LINE__, (msg), cond, errnum, true)
/**
   @brief teste errnum et affiche une erreur si errnum est un code d'erreur différent de 0
*/
#define warning_error(msg, cond, errnum) __error_(__FILE__, __LINE__, (msg), cond, errnum, false)

/**
   @brief fonction générique de la gestion d'erreur
   @param file,ligne : le nom et la ligne du fichier où a lieu l'erreur
   @param msg : le message à afficher en cas d'erreur
   @param cond : la condition (l'erreure est déclanchée si elle est vraie)
   @param errnum : code d'erreur système
   @param exception : si vrai envoie une exception en cas d'erreur,
   sinon, c'est just un warning
*/
void
__error_(const char *file, const int ligne, const std::string &msg, bool cond, int errnum, bool exception) {
    
  if (cond) {
    // il y a bien une erreur
    std::ostringstream c;
    if (exception) {
      c <<  "\033[31;11mERREUR\033[00m";
    } else {
      c <<  "\033[34;11mWARNING\033[00m";
    }
    c << " fichier " << file << " ligne " <<  ligne << " : ";
    if (errnum != 0) {
      std::system_error err = std::system_error(errno, std::system_category(), msg);
      c << err.what() << std::endl;
      std::cerr << c.str();
      if (exception) {
        throw err;
      }
    } else {
      std::runtime_error err = std::runtime_error(msg);
      c << err.what() << std::endl;
      std::cerr << c.str();
      if (exception) {
        throw err;
      }
    }
  }
}

//########################################
//##  Solution lecture/ecriture
//########################################

std::vector<char> lire_donnees(int fd, int taille) {
  std::vector<char> res(taille);

  int deja = 0;
  while(deja < taille) {
    int nb = read(fd, res.data()+deja, taille-deja);
    exit_error("read", nb == -1, errno);
    
    if (nb == 0) {
      std::cerr << "Problème le flux est fermé alors qu'il reste "
		<< taille-deja << " octets à lire"  << std::endl;
      // je vide le contenu du tableau
      memset(res.data()+deja, 0, taille-deja);
      break;
    }

    deja += nb;
  }
  
  return res;
}

void envoie_donnees(int fd, std::vector<char> data) {
  int deja = 0;
  int taille = data.size();
  
  while (deja < taille) {
    int nb = write(fd, data.data()+deja, taille-deja);
    exit_error("write", nb == -1, errno);
    
    deja += nb;
  }
}

//#########################################
//##  solution election
//#########################################


int election(int num, int nb_proc, int entree, int sortie) {
  int leader = num;
  std::ostringstream c;
  std::vector<char> data(32);

  srand((num+1)*time(NULL));
  int val = rand()%100;

  c.str("");
  c << "Processus " << num
    << " ma valeur est " << val << std::endl;
  std::cerr << c.str();

  // std::string osf;
  // getline(std::cin, osf);
  
  if (num != 0) {
    std::vector<char> mess = lire_donnees(entree, 16);
    int val_autre = atoi(mess.data());
    mess = lire_donnees(entree, 16);
    int num_autre = atoi(mess.data());

    if (val_autre>=val) {
      val = val_autre;
      leader = num_autre;
    }

    c.str("");
    c << "Processus " << num
      << " recu " << val_autre
      << " ma valeur est " << val
      << " le leader est " << leader 
      << std::endl;
    std::cerr << c.str();

  }

  snprintf(data.data(), 32, "%15d %15d", val, leader);
  envoie_donnees(sortie, data);
  
  std::vector<char> mess = lire_donnees(entree, 16);
  int val_autre = atoi(mess.data());
  mess = lire_donnees(entree, 16);
  int num_autre = atoi(mess.data());
  
  if (val_autre>=val) {
    val = val_autre;
    leader = num_autre;
  }
  c.str("");
  c << "Processus " << num
    << " recu " << val_autre
    << " ma valeur est " << val
    << " le leader est " << leader 
    << std::endl;
  std::cerr << c.str();

  if (num != nb_proc-1) {
    snprintf(data.data(), 33, "%15d%16d", val, leader);
    envoie_donnees(sortie, data);
  }
  
  return leader;
}


//########################################
//##  Fonctions pour tester
//########################################

void usage(const std::string &nomprog) {
  std::cerr << "Usage "
            << nomprog
            << " [-n <nb de processus>] [-s] "
            << std::endl;
}


int main(int argc, char* argv[]) {
  int i;
  int option;
  std::ostringstream c;
  bool signaux = false;
  int nb_procs = 5;

  // Lecture des options
  while ((option = getopt(argc, argv, "sn:")) != -1) {
    switch(option) {
    case 'n': { // le nombre de processus
      nb_procs = atoi(optarg);
      if (nb_procs == 0) {
        std::ostringstream c;
        c << "!!! le nombre de processus " << optarg
          << " n'est pas correcte !!!";
        std::cerr << c.str() << std::endl;
        usage(argv[0]);
        exit_error(c.str().c_str(), true, 0);
      }
      break;
    }
    case 's': { // on demande de gérer tout avec des signaux
      signaux = true;
      break;
    }
    default :
      std::ostringstream c;
      c << "!!! l'option " << optopt
	<< " est inconnue ou mal utilisée !!!";
      std::cerr << c.str() << std::endl;
      usage(argv[0]);
      exit_error(c.str().c_str(), true, 0);
    }
  }

  std::cerr << "lancement de l'élection avec "
	    << nb_procs << " processus" << std::endl;
  if (signaux) {
    std::cerr << "Les signaux ne sont pas encore gérés" << std::endl; 
  }

  std::vector<int> entrees(nb_procs);
  std::vector<int> sorties(nb_procs);
  for (i=0; i<nb_procs; i++) {
    int p[2];
    int r = pipe(p);
    exit_error("pipe", r == -1, errno);

    entrees[i] = p[0];
    sorties[i] = p[1];
  }

  pid_t pid;
  std::vector<pid_t> fils;
  for (i=0; i<nb_procs; i++) {
    pid = fork();
    exit_error("Problème de fork", pid==-1, errno);

    if (pid == 0) {// c'est le fils
      // c << "break " << i << std::endl;
      // std::cerr << c.str();
      break;
    } else {
      fils.push_back(pid);
    }
  }
  int num = i;  
  
  if (pid == 0) { // c'est un fils
    c.str("");
    c << "Je suis le fils " << num << std::endl;
    std::cerr << c.str();
    
    // On reconnait les 2 pipes d'entrée et de sortie pour ce proc
    int entree_g;
    int sortie_d;
    if (num == 0) {
      sortie_d = sorties[0];
      entree_g = entrees[nb_procs-1];
    } else {
      sortie_d = sorties[num];
      entree_g = entrees[num-1];
    }

    // on ferme tout les pipes sauf ceux qui sont utilisés
    for (i=0; i<nb_procs; i++) {
      if (sorties[i] != sortie_d) {
	int r = close(sorties[i]);
	exit_error("close", r==-1, errno);
      }
      if (entrees[i] != entree_g) {
	int r = close(entrees[i]);
	exit_error("close", r==-1, errno);
      }
    }

    // c.str("");
    // c << "Processus " << num
    //   << " entree " << entree_g
    //   << " sortie " << sortie_d
    //   << std::endl;
    // std::cerr << c.str();
    
    int leader = election(num, nb_procs, entree_g, sortie_d);

    std::cerr << "Processus " << num
	      << " le leader est " << leader << std::endl;

    return leader;
  } else { // je suis le père

    for (i=0; i<nb_procs; i++) {
      int r;
      r = close(sorties[i]);
      exit_error("close", r==-1, errno);
      r = close(entrees[i]);
      exit_error("close", r==-1, errno);
    }

    std::vector<int> leaders; 
    for (i=0; i < (int)fils.size(); i++) {
      int status;
      int r = waitpid(fils[i], &status, 0);
      exit_error("waitpid", r==-1, errno);

      exit_error("Le processus ne s'est pas terminé correctement", !WIFEXITED(status), 0);

      leaders.push_back(WEXITSTATUS(status));
    }


    int leader = leaders[0];
    int i=0;
    for (auto lead : leaders) {
      if (lead != leader) {
	std::cerr << "Attention le leader n'est pas le même pour le processus 0 et pour " << i << std::endl;
	exit(1);
      }
      i++;
    }

    std::cout << "Pour tous les processus, le leader est " << leader << std::endl;
  }

  return 0;
}
